import 'dart:convert';
import 'dart:ui' as ui;

import 'package:bakalarka_frontend/bloc/eraser_cubit.dart';
import 'package:bakalarka_frontend/bloc/offset_with_stroke_cubit.dart';
import 'package:bakalarka_frontend/bloc/stroke_cubit.dart';
import 'package:bakalarka_frontend/widgets/annotation.dart';
import 'package:bakalarka_frontend/widgets/training.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:http/http.dart';
import 'package:image/image.dart' as img;
import 'package:localstorage/localstorage.dart';

class Iteration extends StatelessWidget {
  final LocalStorage _storage = LocalStorage('info');
  final String _datasetName;

  Iteration(this._datasetName);

  @override
  Widget build(BuildContext context) {
    Map<String, dynamic> datasets = _storage.getItem('datasets') == null
        ? Map()
        : jsonDecode(jsonEncode(_storage.getItem('datasets')));
    Map<String, dynamic> iterations = _storage.getItem('iterations') == null
        ? Map()
        : jsonDecode(jsonEncode(_storage.getItem('iterations')));

    return MultiBlocProvider(
      providers: [
        BlocProvider<OffsetWithStrokeCubit>(
            create: (_) => OffsetWithStrokeCubit()),
        BlocProvider<StrokeCubit>(create: (_) => StrokeCubit()),
        BlocProvider<EraserCubit>(create: (_) => EraserCubit()),
      ],
      child: Builder(builder: (context) {
        var body, button;
        if (iterations[_datasetName]['manual_annotation'] == 0) {
          body = Training(_datasetName);
          button = null;
        } else {
          var image;
          for (var i in datasets[_datasetName]['images']) {
            if (datasets[_datasetName]['annotated'] == null ||
                !datasets[_datasetName]['annotated'].contains(i)) {
              image = i;
              break;
            }
          }
          var imageinbase64 =
              post(Uri.parse('http://localhost:8000/download.php'), body: {
            'dataset': _datasetName,
            'name': image,
          }, headers: {
            "Access-Control-Allow-Origin": "*"
          });
          if (datasets[_datasetName]['preannotated'] != null &&
              datasets[_datasetName]['preannotated'].contains(image)) {
            drawPremask(image, context);
          }
          body = FutureBuilder(
            future: imageinbase64,
            builder: (context, snapshot) {
              if (snapshot.hasData) {
                Response response = snapshot.data as Response;
                return Annotation(response.body);
              }
              return Center(child: CircularProgressIndicator());
            },
          );
          button = FloatingActionButton.extended(
            onPressed: () async {
              await finishAnnotatingImage(
                  context, imageinbase64, image, datasets, iterations);
            },
            label: Text(iterations[_datasetName]['manual_annotation'] == 1
                ? 'Finish'
                : 'Next'),
            icon: Icon(Icons.done),
          );
        }
        return Scaffold(
          appBar: AppBar(
            title: Row(
              mainAxisAlignment: MainAxisAlignment.spaceBetween,
              children: [
                Text('Iteration/annotation'),
                Text('Dataset: ' + _datasetName),
                Text(iterations[_datasetName]['manual_annotation'].toString() +
                    ' images left to annotate'),
              ],
            ),
          ),
          body: body,
          floatingActionButton: button,
        );
      }),
    );
  }

  Future<void> finishAnnotatingImage(
      BuildContext context,
      Future<Response> imageinbase64,
      image,
      Map<String, dynamic> datasets,
      Map<String, dynamic> iterations) async {
    String mask = await createMask(context, imageinbase64);

    post(Uri.parse('http://localhost:8000/upload.php'), body: {
      'dataset': _datasetName,
      'name': '.mask' + image.split('.')[0] + '.png',
      'image': mask,
    }, headers: {
      "Access-Control-Allow-Origin": "*"
    });

    if (datasets[_datasetName]['annotated'] == null)
      datasets[_datasetName]['annotated'] = [];
    datasets[_datasetName]['annotated'].add(image);
    if (datasets[_datasetName]['preannotated'] != null)
      datasets[_datasetName]['preannotated'].remove(image);
    _storage.setItem('datasets', datasets);
    iterations[_datasetName]['manual_annotation'] -= 1;
    _storage.setItem('iterations', iterations);
    Navigator.of(context).pop();
    Navigator.of(context)
        .push(MaterialPageRoute(builder: (context) => Iteration(_datasetName)));
  }

  Future<String> createMask(
      BuildContext context, Future<Response> imageinbase64) async {
    final recorder = ui.PictureRecorder();
    final canvas = Canvas(recorder);

    Paint brush = Paint();
    brush.color = Colors.white;
    brush.strokeCap = StrokeCap.round;
    brush.isAntiAlias = true;

    List<OffsetWithStroke?> points = [];
    if (BlocProvider.of<OffsetWithStrokeCubit>(context).state.isNotEmpty)
      points =
          BlocProvider.of<OffsetWithStrokeCubit>(context).state.last.toList();

    var currentImage = Image.memory(
      base64Decode((await imageinbase64).body),
    );

    var sizes = await widthHeight(currentImage);

    double widthRatio =
        ((MediaQuery.of(context).size.width - 50) * 0.8) / sizes[0].toDouble();
    double heightRatio =
        ((MediaQuery.of(context).size.height - 50) * 0.8) / sizes[1].toDouble();
    double multiplier = widthRatio < heightRatio ? widthRatio : heightRatio;

    for (int i = 0; i < points.length - 1; i++) {
      if (points[i] != null && points[i + 1] != null) {
        brush.strokeWidth = points[i]!.stroke / multiplier;
        brush.blendMode =
            points[i]!.isEraser ? BlendMode.clear : BlendMode.color;
        canvas.drawLine(points[i]!.offset / multiplier,
            points[i + 1]!.offset / multiplier, brush);
      } else if (points[i] != null) {
        brush.strokeWidth = points[i]!.stroke / multiplier;
        brush.blendMode =
            points[i]!.isEraser ? BlendMode.clear : BlendMode.color;
        canvas.drawPoints(
            ui.PointMode.points, [points[i]!.offset / multiplier], brush);
      } else if (points[i + 1] != null) {
        brush.strokeWidth = points[i + 1]!.stroke / multiplier;
        brush.blendMode =
            points[i + 1]!.isEraser ? BlendMode.clear : BlendMode.color;
        canvas.drawPoints(
            ui.PointMode.points, [points[i + 1]!.offset / multiplier], brush);
      }
    }

    final picture = recorder.endRecording();
    final img = await picture.toImage(sizes[0], sizes[1]);
    var mask = base64Encode(
        (await img.toByteData(format: ui.ImageByteFormat.png))!
            .buffer
            .asUint8List());
    return mask;
  }

  void drawPremask(image, BuildContext context) {
    post(Uri.parse('http://localhost:8000/download.php'), body: {
      'dataset': _datasetName,
      'name': '.premask' + image.split('.')[0] + '.png',
    }, headers: {
      "Access-Control-Allow-Origin": "*"
    }).then((value) {
      img.Image? mask = img.decodeImage(base64Decode(value.body));
      if (mask != null) {
        double widthRatio = ((MediaQuery.of(context).size.width - 50) * 0.8) /
            mask.width.toDouble();
        double heightRatio = ((MediaQuery.of(context).size.height - 50) * 0.8) /
            mask.height.toDouble();
        double multiplier = widthRatio < heightRatio ? widthRatio : heightRatio;
        var bytes = mask.getBytes(format: img.Format.luminance);
        BlocProvider.of<OffsetWithStrokeCubit>(context).newVersion();
        List<OffsetWithStroke?> points = [];
        for (int x = 0; x < mask.width; x++) {
          for (int y = 0; y < mask.height; y++) {
            if (bytes[y * mask.width + x] == 255) {
              points.add(OffsetWithStroke(
                  Offset(x.toDouble(), y.toDouble()) * multiplier,
                  multiplier,
                  false,
                  cap: StrokeCap.square));
            } else if (points.isNotEmpty && points.last != null) {
              points.add(null);
            }
          }
        }
        BlocProvider.of<OffsetWithStrokeCubit>(context).add(points);
      }
    });
  }
}
